import sys
import os
from bvareader import reader
from bvareader.preprocessing import add_rotation
from bvareader.preprocessing import preprocess_bva_data


# ' The command needs to be in the following format:
# ' path to the file originating in the active directory, using double quotes and os.sensitive separator
# ' Name of the output file without extension, e.g. "patient154"
# '
# ' Example: python xml_to_csv.py "example_data\p154_456.xml" "patient154"

def bva_preprocess_xml():
    # Validations
    path = get_sys_filepath(sys.argv)
    output_name = default_sys_argument(sys.argv)

    pd_bva = reader.read_xml_bva(path)
    pd_bva = add_rotation(pd_bva)
    pd_bva2 = preprocess_bva_data(pd_bva)
    pd_sync = reader.read_xml_sync(path)
    pd_phases = reader.read_xml_phases(path)
    pd_start_stop = reader.read_measure_start_stop(path)

    reader.save_csv(pd_bva, output_name + 'positions_full.csv')
    reader.save_csv(pd_bva2, output_name + 'positions.csv')
    reader.save_csv(pd_sync, output_name + 'sync_times.csv')
    reader.save_csv(pd_phases, output_name + 'phases.csv')
    reader.save_csv(pd_start_stop, output_name + 'measure_start_stop.csv')


def bva_positions_table():
    # Validations
    path = get_sys_filepath(sys.argv)
    output_name = default_sys_argument(sys.argv) + 'positions'

    pd_bva = reader.read_xml_bva(path)
    pd_bva = add_rotation(pd_bva)
    pd_bva2 = preprocess_bva_data(pd_bva)

    reader.save_csv(pd_bva, output_name + '_full.csv')
    reader.save_csv(pd_bva2, output_name + '.csv')


def bva_sync_times_table():
    # Validations
    path = get_sys_filepath(sys.argv)
    output_name = default_sys_argument(sys.argv) + 'sync_times.csv'
    pd_sync = reader.read_xml_sync(path)
    reader.save_csv(pd_sync, output_name)


def bva_phases_table():
    # Validations
    path = get_sys_filepath(sys.argv)
    output_name = default_sys_argument(sys.argv) + 'phases.csv'
    pd_phases = reader.read_xml_phases(path)
    reader.save_csv(pd_phases, output_name)


def bva_measure_start_stop_table():
    # Validations
    path = get_sys_filepath(sys.argv)
    output_name = default_sys_argument(sys.argv) + 'measure_start_stop.csv'
    pd_start_stop = reader.read_measure_start_stop(path)
    reader.save_csv(pd_start_stop, output_name)


def xml_settings_to_csv():
    # Validations
    if (len(sys.argv) < 2):
        sys.exit('You have to provide xml input file')
    else:
        # TODO - issue with comman with single instead of double quotes
        path = os.getcwd() + '\\' + sys.argv[1]

    if (len(sys.argv) < 3):
        print('output file will be settings.csv')
        output = 'settings'
    else:
        output = sys.argv[2]

    pd_settings = reader.read_xml_settings(path)

    reader.save_csv(pd_settings, output + '.csv')


def get_sys_filepath(args):
    if (len(args) < 2):
        sys.exit('You have to provide valid input file path')
    else:
        # TODO - issue with comman with single instead of double quotes
        path = os.getcwd() + '\\' + args[1]
        # check existence
    return(path)


def default_sys_argument(args):
    if (len(args) < 3): 
        output_name = ''
    else:
        output_name = args[2] + '_'
    return(output_name)



from bvareader import reader
import pandas as pd


def test_loading_bva_xml(bva_xml_data_path):
    pd_bva = reader.read_xml_bva(bva_xml_data_path)
    assert isinstance(pd_bva, pd.core.frame.DataFrame)


def test_loading_sync_csv(sync_csv_data_path):
    pd_sync = reader.read_sync_file(sync_csv_data_path)
    assert isinstance(pd_sync, pd.core.frame.DataFrame)


def test_loading_settings_xml(settings_xml_data_path):
    pd_settings = reader.read_xml_settings(settings_xml_data_path)
    assert isinstance(pd_settings, pd.core.frame.DataFrame)


def test_loading_measure_start(settings_xml_data_path):
    pd_start_stop = reader.read_measure_start_stop(settings_xml_data_path)
    assert isinstance(pd_start_stop, pd.core.frame.DataFrame)



import pytest

@pytest.fixture
def bva_xml_data_path():
    return "tests/test_data/example_bva.xml"


@pytest.fixture
def sync_csv_data_path():
    return "tests/test_data/example_sync.csv"


@pytest.fixture
def settings_xml_data_path():
    return 'tests/test_data/example_test_settings.xml'



import numpy as np


# Wrapper around multiple preprocessing functions
# Adds rotation
# clears out positions
def preprocess_bva_data(pd_bva):
    pd_bva = add_rotation(pd_bva)
    pd_bva = add_midpoint(pd_bva)
    pd_bva = remove_unnecessary_columns(pd_bva)
    pd_bva = rename_columns(pd_bva)
    # Needs to be after column renaming because it removes position_x column which is renamed above
    pd_bva = clear_out_of_arena_positions(pd_bva)
    return(pd_bva)


# Removes any points that are out of constraints of the bva and replaces then with NAs
def clear_out_of_arena_positions(pd_bva):
    pd_bva.position_x[np.abs(pd_bva.position_x) > 250] = np.nan
    pd_bva.position_x[np.abs(pd_bva.position_x) > 250] = np.nan
    return(pd_bva)


def add_rotation(pd_bva):
    x_cross, y_cross = calculate_perpendicular_cross(pd_bva.Left_x, pd_bva.Left_y, 
                                                     pd_bva.Right_x, pd_bva.Right_y, pd_bva.Front_x, pd_bva.Front_y)
    front_x, front_y = pd_bva.Front_x - x_cross, pd_bva.Front_y - y_cross
    zipped = list(zip(front_x, front_y))
    pd_bva['rotation_x'] = angle_between(zipped, [(0, 1)])
    return(pd_bva)


def add_midpoint(pd_bva):
    pd_bva['midpoint_x'] = (pd_bva.Right_x + pd_bva.Left_x + pd_bva.Front_x) / 3
    pd_bva['midpoint_y'] = (pd_bva.Right_y + pd_bva.Left_y + pd_bva.Front_y) / 3
    return pd_bva


# Removes columns used only for calculation of rotation
def remove_unnecessary_columns(pd_bva, force=False):
    # checks if rotation has been calculated
    cols = ['Point_x', 'Point_y', 'Right_x', 'Right_y', 'Left_x', 'Left_y', 'Front_x', 'Front_y', 'timestamp_bva']
    if 'rotation_x' not in pd_bva.columns:
        Warning('You are deleting columns without calculating rotation first. set force to True if you want to really delete')
    if 'rotation_x' in pd_bva.columns or force:
        pd_bva = pd_bva.drop(cols, axis=1)
    return(pd_bva)


def rename_columns(pd_bva):
    pd_bva = pd_bva.rename(columns={"midpoint_x": 'position_x', 'midpoint_y': 'position_y'})
    return(pd_bva)


# Calculates the perpendicular line to left and right point line which goes through front
def calculate_perpendicular_cross(line1_x, line1_y, line2_x, line2_y, origin_x, origin_y):
    k = (((line2_y-line1_y) * (origin_x-line1_x)) - ((line2_x-line1_x) * (origin_y-line1_y))) / (np.square(line2_y-line1_y) + np.square(line2_x-line1_x))
    x4 = origin_x - (k * (line2_y-line1_y))
    y4 = origin_y + (k * (line2_x-line1_x))
    return(x4, y4)


def calculate_perpendicular_cross_classical(line1_x, line1_y, line2_x, line2_y, origin_x, origin_y):
    slope = (line2_x - line1_x) / (line2_y - line1_y)
    b = line1_y + line1_x * slope
    perp_slope = (line1_y - line2_y) / (line2_x - line1_x)
    perp_b = origin_y + (perp_slope * origin_x)  # linear coef B
    cross_x = (perp_b - b)/(slope-perp_slope)
    cross_y = perp_slope*cross_x + perp_b
    return(cross_x, cross_y)


# p1 are lists of touples [(0,1),(1,0)]
def angle_between(p1, p2):
    x, y = zip(*p1[::1])
    ang1 = np.arctan2(x, y)
    x, y = zip(*p2[::1])
    ang2 = np.arctan2(x, y)
    degrees = np.rad2deg((ang1 - ang2) % (2 * np.pi))
    return(degrees)



from collections import defaultdict


def flatten_list(list_of_lists):
    return([item for sublist in list_of_lists for item in sublist])


def find_duplicates(mylist):
    # https://stackoverflow.com/questions/11236006/identify-duplicate-values-in-a-list-in-python
    D = defaultdict(list)
    for i, item in enumerate(mylist):
        D[item].append(i)
    D = {k: v for k, v in D.items() if len(v) > 1}
    # returns hash like {20: [0, 3], 30: [1, 4]}
    i_duplicates = []
    for k in D.keys():
        i_duplicates += D[k][1:]
    return i_duplicates


def remove_at_indices(mylist, indices):
    return [i for j, i in enumerate(mylist) if j not in indices]



import matplotlib.pyplot as plt
from bvareader.preprocessing import calculate_perpendicular_cross


def plot_walking(pd_bva):
    plt.plot(pd_bva.Point_x, pd_bva.Point_y)
    plt.show()


def plot_position_histograms(pd_bva):
    fig, axes = plt.subplots(2)
    pd_bva.hist(column='Point_x', bins=100, ax=0)
    pd_bva.hist(column='Point_y', bins=100, ax=1)


def plot_triangle(pd_bva, index):
    line = pd_bva.iloc[index]
    plot_triangle_height(line.Left_x, line.Left_y, line.Right_x, line.Right_y, line.Front_x, line.Front_y)


def plot_triangle_height(x1, y1, x2, y2, top_x, top_y):
    plt.scatter(top_x, top_y, marker='o', s=10)
    plt.scatter(x1, y1, marker='o', s=10)
    plt.scatter(x2, y2, marker='o', s=10)
    cross_x, cross_y = calculate_perpendicular_cross(x1, y1, x2, y2, top_x, top_y)
    plt.scatter(cross_x, cross_y, marker='o', s=10)
    plt.plot([cross_x, top_x], [cross_y, top_y], lw=1)
    plt.plot([x1, x2], [y1, y2], lw=1)
    plt.gca().set_aspect('equal', adjustable='box')
    plt.show()


import pandas as pd
import xml.etree.ElementTree as ET
from datetime import datetime
from bvareader.helpers import flatten_list
from bvareader.helpers import find_duplicates
from bvareader.helpers import remove_at_indices


def read_xml_phases(path):
    root = ET.parse(path).getroot()
    phase_times = []
    i_phase = 1
    for phase in root.iter('Phase'):
        row = [str(i_phase)]
        # phase_time = 0
        tps = phase.findall("./MousePath/TimestampPoint")
        if len(tps) == 0:
            row += [float('nan'), float('nan'), float('nan')]
        else:
            row.append(real_timestamp(tps[0]))
            row.append(real_timestamp(tps[-1]))
            row.append(float(tps[-1].find('Timestamp').text))
        phase_times.append(row)
        i_phase += 1
    pd_phases = pd.DataFrame(phase_times, columns=["phase_number", "timestamp_start", 
                             "timestamp_end", "claimed_length"])
    return(pd_phases)


def read_xml_sync(path):
    root = ET.parse(path).getroot()
    times = []
    for phase in root.iter('Phase'):
        for sync in phase.iter('SyncEEGAction'):
            times.append(real_timestamp(sync))
    pd_times = pd.DataFrame(data={'order': list(range(1, len(times)+1)), 'timestamp': times})
    return(pd_times)


def read_xml_bva(path):
    root = ET.parse(path).getroot()
    POINTS = ['Point', 'Front', 'Left', 'Right']
    bva_mat = []
    continuous_time = 0
    for phase in root.iter('Phase'):
        for TimestampPoint in phase.iter('TimestampPoint'):
            # old points had only Timestamp
            phase_time = float(TimestampPoint.find('Timestamp').text)
            phase_datetime_real = real_timestamp(TimestampPoint)
            row = []
            row.append(phase_time + continuous_time)
            row.append(phase_datetime_real)
            for point in POINTS:
                xy = TimestampPoint.find(point)
                if xy is not None:
                    row.append(float(xy.find('X').text))
                    row.append(float(xy.find('Y').text))
                else:
                    row.append(float("NaN"))
                    row.append(float("NaN"))
            bva_mat.append(row)
        continuous_time += phase_time
    # adds colnames for all the points
    colnames = ['timestamp_bva', 'timestamp'] + (flatten_list([[x + "_x", x + "_y"] for x in POINTS]))
    pd_bva = pd.DataFrame(bva_mat, columns=colnames)
    return(pd_bva)


def read_measure_start_stop(path):
    root = ET.parse(path).getroot()
    start_times, stop_times, i_phases = [], [], []
    i_phase = 0 # raised to 1 in the first phase
    for phase in root.iter('Phase'):
        i_phase += 1
        measures_starts = phase.findall("./MousePath/MeasureStartItem")
        measures_stops = phase.findall("./MousePath/MeasureStopItem")
        if(len(measures_starts) != len(measures_stops)):
            raise Exception('there is unequal number of start measures and stop measures')
        if(len(measures_starts) < 1):
            continue
        phase_timestamp = real_timestamp(phase.find("./MousePath/TimestampPoint"))
        for n in range(0, len(measures_starts)):
            start_times += [float(measures_starts[n].find("Timestamp").text) + phase_timestamp]
            stop_times += [float(measures_stops[n].find("Timestamp").text) + phase_timestamp]
            i_phases += [str(i_phase)]
    pd_start_stops = pd.DataFrame(data={'measure_start': start_times, 'measure_stop': stop_times,
                                        'phase_number': i_phases})
    return pd_start_stops


def read_xml_settings(path):
    root = ET.parse(path).getroot()
    pd_settings = pd.DataFrame()
    i_phase = 1
    for phase in root.iter('phase'):
        keys, values = element_to_row(phase)
        keys += ["phase_number"]
        values += [str(i_phase)]
        # Check if the keys are still the same and values of the same length
        pd_row = pd.DataFrame([values], columns=keys)
        pd_settings = pd.concat([pd_settings, pd_row])
        i_phase += 1
    return pd_settings


def read_sync_file(path):
    pd_sync = pd.read_csv(path, sep=",")
    pd_sync.time = (pd.to_datetime(pd_sync.time, format='%H:%M:%S') -
                    pd.to_datetime("00:00:00", format='%H:%M:%S')).dt.total_seconds()
    pd_sync.time = pd_sync.time + pd_sync.ms / 1000
    return(pd_sync)


# ' Heavily needs description
def element_to_row(element):
    keys = flatten_list([[element.tag + "_" + x] for x in list(element.attrib.keys())])
    values = list(element.attrib.values())
    for el in element:
        tag = el.tag
        tag_keys = list(el.attrib.keys())
        keys += flatten_list([[tag + "_" + x] for x in tag_keys])
        tag_values = list(el.attrib.values())
        values += tag_values
        value = el.text
        if value is not None:
            keys += [tag + "_value"]
            values += [str(value)]
        children_keys, child_values = element_to_row(el)
        keys, values = keys + children_keys, values + child_values
    i_duplicate = find_duplicates(keys)
    keys, values = remove_at_indices(keys, i_duplicate), remove_at_indices(values, i_duplicate)
    return keys, values


# ' Converts XML BVA timestamp to POSIX timestamp
def real_timestamp(element):
    # timestamp real is in form of 01/29/2019 10:02:45.574
    dt = datetime.strptime(element.find('TimestampReal').text,
                           '%m/%d/%Y %H:%M:%S.%f')
    return(float(dt.timestamp()))


# ' Wrapper around 
def save_csv(pd_bva, path, dec_points=4):
    f = '%.'+str(dec_points)+'f'
    pd_bva.to_csv(path, sep=";", index=False, float_format=f)



import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="bvareader",
    version="0.1.3",
    author="Lukáš Hejtmy Hejtmánek",
    author_email="hejtmy@gmail.com",
    description="Package to read and process BVA data from LF Motol",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hejtmy/bva-reader",
    packages=setuptools.find_packages(),
    install_requires=[
      'pandas',
      'numpy>=1.16.0',
      'matplotlib',
      ],
    tests_require=["pytest", ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
    entry_points={
      'console_scripts': [
        'bva-preprocess-xml = bvareader.commands:bva_preprocess_xml',
        'bva-positions-table = bvareader.commands:bva_positions_table',
        'bva-phases-table = bvareader.commands:bva_phases_table',
        'bva-measures-start-stop-table = bvareader.commands:bva_measure_start_stop_table',
        'bva-sync-times-table = bvareader.commands:bva_sync_times_table',
        'bva-settings-to-csv = bvareader.commands:xml_settings_to_csv',
        ]
    }
)